| $Id: updown.asm,v 1.6 2001/08/25 21:38:49 dick Exp $
|======================================================================|
| updown.asm - Nokia Mediamaster subroutine to up/down-load RAM/Flash
|======================================================================|
|
| This subroutine is used to implement the memory download, memory
| upload, and firmware upgrade options of the 'mmlink' utility.
| First, a '+' character is sent to the PC, to signal that we are
| ready to accept a command. A command consists of a command character
| followed by a begin and an end address, transmitted as 4-byte big
| endian numbers. The command is followed by a CRC16 checksum over these
| nine characters. The communication during this command phase goes like
| this:
|
|   PC: start this subroutine with CBEG D
|   MM: send '+' to signal that it is ready
|   PC: send command character
|   PC: send begin address as big endian 4-byte number
|   PC: send end address as big endian 4-byte number
|   PC: send CRC16 over last 9 bytes
|   MM: send '+' to acknowledge the command
|
| In the data transfer phase, the data is sent in chunks of 1k, with
| an acknowledge after each block. A CRC16 checksum is used to protect
| against communication errors. The following commands are currently
| implemented:
|
| D - Memory download
| -------------------
| Download a memory region from the Mediamaster to the PC:
|   PC: send 'd'
|   MM: send 1k data
|   PC: send 'd'
|   MM: send 1k data (or less than 1k for the last block)
|   ...
|   PC: send 'd'
|   MM: send 2 bytes with CRC-16 checksum
|   PC: send 'd'
|
| U - RAM memory upload
| ---------------------
| Upload data to the RAM memory of the Mediamaster:
|   MM: send '+'
|   PC: send 1k data
|   MM: send '+'
|   PC: send 1k data (or less than 1k for the last block)
|   ...
|   MM: send '+'
|   PC: send 2 bytes with CRC-16 checksum
|   MM: send '+'
|
| F - Flash memory upload
| -----------------------
| A flash memory upload is different from a RAM upload because the
| data is first read into a RAM buffer. Although the CRC is checked
| during the upload into the RAM buffer, it is checked another time
| to make sure the RAM is not overwritten. When the 16-bit CRC is OK,
| the buffer is copied to flash memory, 64k at a time. For each 64k
| block, the flash is erased, and the data is programmed into the flash.
| At the end, the flash memory is verified by calculating the CRC again.
| Buffering the data in RAM will eliminate the risk of ending up with
| an incomplete firmware because of communication problems. But of
| course, there is always the risk of a power failure ...
| The communication between the PC and the Mediamaster is the same
| as for a memory upload, with the addition of:
|   MM: send '+' when the CRC checksum of the RAM buffer is OK
|   MM: send '.' before a sector is flashed, to prevent a timeout
|   MM: send '+' when the writing is finished
|   MM: send '+' when the flash verify was OK
| At the end of a flash memory upload, the Mediamaster will be RESET.
|
| T - Download Teletext font
| --------------------------
| Find the Teletext font and send it to the PC. The begin and end
| addresses are ignored in this case. The position of the font table
| is determined by searching the flash memory. When the font position
| is found, the Mediamaster sends an acknowledge:
|   MM: send '+'
| After that, the protocol continues as a normal memory download (D).
|
| Y - Download system font
| ------------------------
| Find the system font and send it to the PC. The begin and end
| addresses are ignored in this case. The position of the font table
| is determined by searching the flash memory. When the font position
| is found, the Mediamaster sends an acknowledge:
|   MM: send '+'
| After that, the protocol continues as a normal memory download (D).
|
| If something's wrong, the Mediamaster will send a '-', followed by
| an error message. A timeout of 5 seconds is used for the serial
| port. This means that this code will abort when it has to wait for
| more than 5 seconds for incoming or outgoing data.
|
| (c) 1999-2001, Dick Streefland
|======================================================================|

|=== RAM buffer used for Flash programming:
	.equ	BUFFER,		0x01000000	| buffer for flash programming
	.equ	BUFEND,		0x010b0000	| >= 0x010b0000 used by DMA?
	.equ	STACKSZ,	512
	.equ	MYSIZE,		(crc_tab-_start+512+STACKSZ+0x3ff)&~0x3ff
	.equ	RELOCA,		(BUFEND-MYSIZE)	| relocation address
	.equ	STACK,		BUFEND		| top of stack
	.equ	BUFSIZ,		(RELOCA-BUFFER)	| maximum size of firmware
|=== The CRC16 polynomal:
	.equ	CRC16,		0x11021		| CRC-CCITT = X^16+X^12+X^5+1
|=== Serial port:
	.equ	IOBASE,		0x00fff000	| I/O base address
	.equ	SER_MR1A,	0x710		| - mode register MR1A
	.equ	SER_SRA,	0x711		| - command register A
	.equ	SER_CRA,	0x712		| - command register A
	.equ	SER_TBA,	0x713		| - transmit buffer A
	.equ	SER_RBA,	0x713		| - receive buffer A
	.equ	SER_OPCR,	0x71d		| - output port control register
	.equ	SER_MR2A,	0x720		| - mode register MR2A
|=== Addresses to control the display:
	.equ	IOPORT,		0x00fff011	| one bit used for display
	.equ	IODVB2000,	0x01000000	| DVB2000s copy of IOPORT value
	.equ	DISPBIT,	2		| bit number for display
	.equ	DISPLAY,	0x03000600	| the display controller
|=== Command bytes for the display controller:
	.equ	CMD_POSITION,	0xa0		| display position (bits 0..2)
	.equ	CMD_DIM,	0xe0		| control dimming (bits 0..2)

|======================================================================|
| Entry point
|======================================================================|
	.global	_start
_start:
	movem.l	a0-a4/d0-d7,-(sp)
	move.w	#0x2700,sr		| disable interrupts
	move.l	sp,a4			| save SP for error exit
	bsr	ser_init		| initialize serial port
	bsr	display_init		| initialize display (init D7)
	moveq	#'+',d0
	bsr	ser_put			| send ACK

	moveq	#-1,d1			| D1 = CRC
	bsr	ser_get			| get command
	bsr	crc
	move.b	d0,d4			| D4 = command byte
	bsr	ser_get4		| get begin address
	bsr	crc4
	move.l	d0,a3			| A3 = begin address
	bsr	ser_get4		| get end address
	bsr	crc4
	move.l	d0,d3			| D3 = end address
	sub.l	a3,d3			| D3 = size
	bsr	ser_get			| get CRC16
	bsr	crc
	bsr	ser_get
	bsr	crc
	lea	(err_crc,pc),a0		| "Bad CRC!"
	tst.w	d1			| CRC16 OK?
	bne.s	error
	moveq	#'+',d0
	bsr	ser_put			| send ACK

	moveq	#-1,d1			| D1 = CRC
	moveq	#0,d2			| D2 = counter
	cmp.b	#'D',d4
	beq	download		| 'D' ==> memory download
	cmp.b	#'U',d4
	beq	upload			| 'U' ==> memory upload
	cmp.b	#'F',d4
	beq	flash			| 'F' ==> flash upload
	cmp.b	#'T',d4
	beq	ttfont			| 'T' ==> teletext font download
	cmp.b	#'Y',d4
	beq	sysfont			| 'Y' ==> system font download
	lea	(err_badcmd,pc),a0	| "Bad cmd!"

|***************************************|
| Error exit
|***************************************|
error:
	moveq	#'-',d0
	bsr	ser_put			| send NACK
	bsr	ser_putstr		| error message to serial port
	bsr	ser_putcrlf
error.ser_put:
	bsr	message			| error message to display
	move.l	a4,sp			| restore SP

|***************************************|
| Normal exit
|***************************************|
exit:
	moveq	#1,d0
	bsr	delay			| wait 0.1 sec for serial output
	cmp.b	#'F',d4			| Flash programming?
	beq.s	reboot			| RESET because RAM is overwritten
	move.w	#0x2000,sr		| enable interrupts again
	movem.l	(sp)+,a0-a4/d0-d7
	rts				| return to DVB2000

|***************************************|
| Reboot
|***************************************|
reboot:
	lea	(msg_reboot,pc),a0
	bsr	message			| "Reboot.."
	reset
	tst.b	0xaaaaaaaa		| causes reset
	bra.s	.

|======================================================================|
| Teletext font download
|======================================================================|

ttfont:
	move.w	#0x2e3c,d0		| "move.l #0x3200,d7"
	move.l	#0x3200,d5
	move.l	#0x3200,d3		| TT font table length
	bra.s	findfont

|======================================================================|
| System font download
|======================================================================|

sysfont:
	move.w	#0x2e3c,d0		| "move.l #0x0fff,d7"
	move.l	#0x0fff,d5
	move.l	#0x1000,d3		| system font table length

|***************************************|
| Find font and send it to the PC
|***************************************|

findfont:
	move.l	#0x20000,a1		| start of DVB2000
	move.l	#0xd0000,a3		| end of DVB2000

font.loop:
	cmp.w	(a1)+,d0
	beq.s	font.match
	cmp.l	a3,a1
	bne.s	font.loop
	lea	(err_notfound,pc),a0	| "NotFound"
	bra.s	error

font.match:
	cmp.l	(a1),d5
	bne.s	font.loop
	cmp.w	#0x49f9,(-8,a1)		| "lea font.l,a4"
	bne.s	font.loop
	move.l	(-6,a1),a3		| A3 = font table address
	moveq	#'+',d0
	bsr	ser_put			| send ACK
	bra.s	download


|======================================================================|
| Memory download
|======================================================================|

down.loop:
	move.l	d2,d0
	and.w	#0x3ff,d0
	bne.s	down.skip
	bsr	progress		| update progress indicator
	bsr	ser_get
	cmp.b	#'d',d0			| wait for ACK
	bne	error
down.skip:
	move.b	(a3,d2.l),d0
	addq.l	#1,d2
	bsr	ser_put
	bsr	crc
download:
	cmp.l	d3,d2
	blo.s	down.loop

	bsr	progress		| update progress indicator
	moveq	#0,d0
	bsr	crc
	bsr	crc
	bsr	ser_get
	cmp.b	#'d',d0			| wait for ACK
	bne	error
	move.w	d1,d0
	rol.w	#8,d0
	bsr	ser_put			| append CRC-16
	rol.w	#8,d0
	bsr	ser_put
	bsr	ser_get
	cmp.b	#'d',d0			| wait for ACK
	bne	error
	bra	exit

|======================================================================|
| Flash memory upload
|======================================================================|

flash:
	lea	(msg_flashing,pc),a0
	bsr	message			| "Flashing"
	cmp.l	#BUFSIZ,d3
	lea	(err_big,pc),a0		| "Too big!"
	bhi	error			| data too large
	lea	(err_noalign,pc),a0	| "No align"
	tst.w	a3			| flash address must be 64k aligned
	bne	error

|***************************************|
| Relocate this code to create a large contiguous RAM buffer
|***************************************|
	lea	(_start,pc),a0
	move.l	#RELOCA,a1
	move.w	#(_end-_start),d0
reloc:
	move.b	(a0)+,(a1)+
	dbra	d0,reloc
	move.l	#STACK,sp
	move.l	sp,a4

	move.l	a3,a1			| A1 = Flash address
	move.l	#BUFFER,a3		| A3 = buffer
	jmp	RELOCA+(upload-_start)

|======================================================================|
| Memory upload
|======================================================================|

up.loop:
	move.l	d2,d0
	and.w	#0x3ff,d0
	bne.s	up.skip
	bsr	progress		| update progress indicator
	moveq	#'+',d0
	bsr	ser_put			| send ACK
up.skip:
	bsr	ser_get
	move.b	d0,(a3,d2.l)
	bsr	crc
	addq.l	#1,d2
upload:
	cmp.l	d2,d3
	bne.s	up.loop

	bsr	progress		| update progress indicator
	moveq	#'+',d0
	bsr	ser_put			| send ACK
	bsr	ser_get			| read 2 CRC bytes
	bsr	crc
	move.b	d0,d5
	lsl.w	#8,d5			| save CRC in D5.w
	bsr	ser_get
	bsr	crc
	move.b	d0,d5
	lea	(err_crc,pc),a0		| "Bad CRC!"
	tst.w	d1
	bne	error			| CRC error
	moveq	#'+',d0
	bsr	ser_put			| send ACK

|***************************************|
| Stop here if this is a normal upload
|***************************************|
	cmp.b	#'F',d4			| Flash programming?
	bne	exit			| no, exit

|***************************************|
| Fill remainder of the last 64k sector with 0xff
|***************************************|
	moveq	#-1,d0
	bra.s	fill.entry
fill.loop:
	move.b	d0,(a3,d2.l)
	addq.l	#1,d2
fill.entry:
	tst.w	d2
	beq.s	fill.done
	cmp.l	#BUFSIZ,d2
	blo.s	fill.loop
fill.done:

|***************************************|
| Check the CRC another time to make sure the buffer is still OK
|***************************************|
	lea	(msg_checking,pc),a0
	bsr	display_string		| "Checking"
	bsr	crc_init
	moveq	#-1,d1			| init CRC
	bsr	crc_fast
	move.w	d5,d0
	bsr	crc2			| add saved CRC
	lea	(err_ram,pc),a0		| "RAM Err!"
	tst.w	d1
	bne	error
	moveq	#'+',d0
	bsr	ser_put			| send ACK
	lea	(msg_crcok,pc),a0
	bsr	message			| "CRC OK"

|***************************************|
| Write to Flash ...
|***************************************|
	lea	(msg_writing,pc),a0
	bsr	message			| "Writing"
	move.l	a3,a0			| A0 = source, A1 = destination
	move.l	d3,d1
	add.l	#0xffff,d1
	swap	d1			| D1.w = number of 64k sectors
	move.l	a1,a3			| A3 = flash start address
	bra.s	fl.entry
fl.loop:
	moveq	#'.',d0
	bsr	ser_put			| prevent timeout
	bsr	erase64k		| erase 64k
	bsr	flash64k		| write 64k into flash
fl.entry:
	dbra	d1,fl.loop
	moveq	#'+',d0
	bsr	ser_put			| send ACK

|***************************************|
| Verify Flash
|***************************************|
	lea	(msg_verify,pc),a0
	bsr	display_string		| "Verify"
	moveq	#-1,d1			| init CRC
	bsr	crc_fast
	move.w	d5,d0
	bsr	crc2			| add saved CRC
	lea	(err_failed,pc),a0	| "Failed!"
	tst.w	d1
	bne	error
	moveq	#'+',d0
	bsr	ser_put			| send final ACK
	lea	(msg_flashok,pc),a0
	bsr.s	message			| "Flash OK"
	bra	exit

|======================================================================|
| Miscellaneous subroutines
|======================================================================|

|***************************************|
| Update progress indicator in display
| D2 = current position
| D3 = total size
|***************************************|
progress:
	movem.l	d0/a0,-(sp)
	subq.l	#8,sp
	lea	(sp),a0
	move.l	d2,d0
	asr.l	#5,d0
	asr.l	#5,d0
	bsr.s	dec3
	move.b	#'/',(a0)+
	move.l	d3,d0
	asr.l	#5,d0
	asr.l	#5,d0
	bsr.s	dec3
	move.b	#'k',(a0)
	lea	(sp),a0
	bsr	display_string		| display counter
	addq.l	#8,sp
	movem.l	(sp)+,d0/a0
	rts

|***************************************|
| Show message in A0 on display & delay 2 seconds
|***************************************|
message:
	move.l	d0,-(sp)
	bsr	display_string		| display message on display
	moveq	#20,d0
	bsr.s	delay			| delay 2 sec
	move.l	(sp)+,d0
	rts

|***************************************|
| Delay 0.1*D0 seconds
|***************************************|
delay:
	movem.l	d0/d1,-(sp)
delay1:
	move.l	#(16000000/10/14),d1
delay2:
	subq.l	#1,d1
	bne.s	delay2
	subq.l	#1,d0
	bne.s	delay1
	movem.l	(sp)+,d0/d1
	rts

|======================================================================|
| Conversions
|======================================================================|

|***************************************|
| Convert D0 to 3-digit decimal number and store it at (a0)+.
|***************************************|
dec3:
	movem.l	d0-d1,-(sp)
	moveq.l	#100,d1
dec3_loop:
	divu.w	d1,d0
	add.b	#'0',d0
	move.b	d0,(a0)+
	asr.l	#8,d0
	asr.l	#8,d0
	divu.w	#10,d1
	bne.s	dec3_loop
	movem.l	(sp)+,d0-d1
	rts

|======================================================================|
| CRC16 calculation
|======================================================================|

|***************************************|
| Add the 2 bytes in in D0.w to CRC in D1.
|***************************************|
crc2:
	rol.w	#8,d0
	bsr.s	crc
	rol.w	#8,d0
	bra.s	crc

|***************************************|
| Add the 4 bytes in in D0 to CRC in D1.
|***************************************|
crc4:
	bsr.w	crc4a
crc4a:
	bsr.w	crc4b
crc4b:
	rol.l	#8,d0
|***************************************|
| Add byte in D0 to CRC in D1.
|***************************************|
crc:
	movem.l	d0/d2,-(sp)
	moveq	#7,d2
crc.loop:
	asl.b	#1,d0
	roxl.w	#1,d1
	bcc.s	crc.0
	eor.w	#(CRC16&0xffff),d1
crc.0:
	dbra	d2,crc.loop
	movem.l	(sp)+,d0/d2
	rts

|***************************************|
| Initialize CRC table for crc_fast
|***************************************|
crc_init:
	movem.l	d0-d2/a0,-(sp)
	lea	(crc_tab,pc),a0
	moveq	#0,d0
	moveq	#0,d2
ci.loop:
	move.w	d2,d1
	bsr.s	crc
	move.w	d1,(a0)+
	add.w	#0x100,d2
	bne.s	ci.loop
	movem.l	(sp)+,d0-d2/a0
	rts

|***************************************|
| Fast CRC16 calculation for a block of memory
| D1 = CRC
| A3 = buffer address
| D3 = buffer size
| Speed: 3.4 sec/MB
|***************************************|
crc_fast:
	movem.l	d0/d2/a0-a1/a3,-(sp)
	lea	(crc_tab,pc),a0
	lea	(a3,d3.l),a1
	moveq.l	#0,d2
	bra.s	cf.entry
cf.loop:
	rol.w	#8,d1
	move.b	d1,d2
	move.b	(a3)+,d1
	move.w	(a0,d2.w*2),d0
	eor.w	d0,d1
cf.entry:
	cmp.l	a1,a3
	blo.s	cf.loop
	movem.l	(sp)+,d0/d2/a0-a1/a3
	rts

|======================================================================|
| Serial port
|======================================================================|

|***************************************|
| Initialize serial port
|***************************************|
ser_init:
	move.l	a0,-(sp)
	move.l	#IOBASE,a0
	move.b	#0xaa,(a0,SER_OPCR)	| output port control
	move.b	#0x93,(a0,SER_MR1A)	| 8 bits, no parity, RxRTS flow control
	move.b	#0x07,(a0,SER_MR2A)	| 1 stop bit, no TxCTS flow control
	move.b	#0x05,(a0,SER_CRA)	| enable RX and TX
	move.l	(sp)+,a0
	rts

|***************************************|
| Output character in D0
|***************************************|
ser_put:
	movem.l	d1/a0,-(sp)
	move.l	#IOBASE,a0
	move.l	#2700000,d1		| timeout: +/- 5 sec
p.wait:
	subq.l	#1,d1
	bmi.s	p.timeout
	btst	#2,(a0,SER_SRA)
	beq.s	p.wait
	move.b	d0,(a0,SER_TBA)		| output character
	movem.l	(sp)+,d1/a0
	rts
p.timeout:
	lea	(err_timeout,pc),a0	| "Timeout!"
	bra	error.ser_put

|***************************************|
| Output string pointed to by A0
|***************************************|
ser_putstr:
	movem.l	d0/a0,-(sp)
	bra.s	ps.entry
ps.loop:
	bsr.s	ser_put
ps.entry:
	move.b	(a0)+,d0
	bne.s	ps.loop
	movem.l	(sp)+,d0/a0
	rts

|***************************************|
| Output CR+LF
|***************************************|
ser_putcrlf:
	move.l	d0,-(sp)
	moveq	#'\r',d0
	bsr.s	ser_put
	moveq	#'\n',d0
	bsr.s	ser_put
	move.l	(sp)+,d0
	rts

|***************************************|
| Output D0 as a hex number
|***************************************|
ser_puthex:
	movem.l	d0/d1/d2,-(sp)
	moveq	#7,d2
	move.l	d0,d1
ph.loop:
	rol.l	#4,d1
	move	d1,d0
	and.b	#0x0f,d0
	cmp.b	#10,d0
	bcs.s	ph.ok
	add.b	#('a'-'9'-1),d0
ph.ok:
	add.b	#'0',d0
	bsr.s	ser_put
	dbra	d2,ph.loop
	movem.l	(sp)+,d0/d1/d2
	rts

|***************************************|
| Read character from input into D0
|***************************************|
ser_get:
	move.l	a0,-(sp)
	move.l	#IOBASE,a0
	move.l	#2700000,d0		| timeout: +/- 5 sec
g.wait:
	subq.l	#1,d0
	bmi.s	g.timeout
	btst.b	#0,(a0,SER_SRA)
	beq.s	g.wait
	moveq	#0,d0
	move.b	(a0,SER_RBA),d0		| read character
	move.l	(sp)+,a0
	rts
g.timeout:
	lea	(err_timeout,pc),a0	| "Timeout!"
	bra	error

|***************************************|
| Read 32-bit number from input into D0
|***************************************|
ser_get4:
	movem.l	d1/d2,-(sp)
	moveq	#0,d1
	moveq	#3,d2
g4.loop:
	bsr.s	ser_get
	rol.l	#8,d1
	or.b	d0,d1
	dbra	d2,g4.loop
	move.l	d1,d0
	movem.l	(sp)+,d1/d2
	rts

|======================================================================|
| Flash programming
|======================================================================|

|***************************************|
| Get Flash ID for memory at (A1) into D0
|***************************************|
|flash_id:
|	movem.l	a3-a4,-(sp)
|	lea	(0xaaa,a1),a3
|	lea	(0x554,a1),a4
|	move.b	#0xaa,(a3)		| first bus cycle
|	move.b	#0x55,(a4)		| second bus cycle
|	move.b	#0x90,(a3)		| ID command
|	moveq	#0,d0
|	move.w	(0,a1),d0		| D0 = vendor
|	swap	d0
|	move.w	(2,a1),d0		| add device code
|	move.b	#0xaa,(a3)		| first bus cycle
|	move.b	#0x55,(a4)		| second bus cycle
|	move.b	#0xf0,(a3)		| RESET command
|	movem.l	(sp)+,a3-a4
|	rts

|***************************************|
| Erase 64k of flash memory.
| A1 = Flash address (64k aligned)
|***************************************|
erase64k:
	movem.l	a3-a4,-(sp)
	lea	(0xaaa,a1),a3
	lea	(0x554,a1),a4
	move.b	#0xaa,(a3)		| first bus cycle
	move.b	#0x55,(a4)		| second bus cycle
	move.b	#0x80,(a3)		| third bus cycle
	move.b	#0xaa,(a3)		| fourth bus cycle
	move.b	#0x55,(a4)		| fifth bus cycle
	move.b	#0x30,(a1)		| erase sector
	move.b	#0x30,(0x4000,a1)	| first 64k of a chip has 4 sectors
	move.b	#0x30,(0x6000,a1)
	move.b	#0x30,(0x8000,a1)
er.wait:
	btst.b	#5,(a1)
	beq.s	er.wait
	move.b	#0xaa,(a3)		| first bus cycle
	move.b	#0x55,(a4)		| second bus cycle
	move.b	#0xf0,(a3)		| RESET command
	movem.l	(sp)+,a3-a4
	rts

|***************************************|
| Write a 64k flash sector.
| A0 = RAM source
| A1 = Flash destination (64k aligned)
|***************************************|
flash64k:
	movem.l	d0-d1/a3-a4,-(sp)
	lea	(0xaaa,a1),a3
	lea	(0x554,a1),a4
	move.w	#0x7fff,d1
floop1:
	move.b	#0xaa,(a3)		| first bus cycle
	move.b	#0x55,(a4)		| second bus cycle
	move.b	#0xa0,(a3)		| third bus cycle
	move.w	(a0)+,d0
	move.w	d0,(a1)
floop2:
	cmp.w	(a1),d0
	bne.s	floop2
	addq.l	#2,a1
	dbra	d1,floop1
	move.b	#0xaa,(a3)		| first bus cycle
	move.b	#0x55,(a4)		| second bus cycle
	move.b	#0xf0,(a3)		| RESET command
	movem.l	(sp)+,d0-d1/a3-a4
	rts

|======================================================================|
| Panel display control
|======================================================================|

|***************************************|
| Initialize the display
|***************************************|
display_init:
	move.l	d0,-(sp)
	move.b	IODVB2000,d7		| D7 = current IOPORT value
	bset.b	#DISPBIT,d7
	move.b	d7,IOPORT
	move.b	#CMD_DIM+1,d0		| set display dimming to 1
	bsr.s	display_cmd
	move.l	(sp)+,d0
	rts

|***************************************|
| Write the command byte in D0 to the display controller
|***************************************|
display_cmd:
	move.l	d1,-(sp)
	bclr.b	#DISPBIT,d7
	move.b	d7,IOPORT
	moveq	#7,d1
dcmd.loop:
	ror.b	#1,d0
	scs	DISPLAY
	dbra	d1,dcmd.loop
	bset.b	#DISPBIT,d7
	move.b	d7,IOPORT
	move.l	(sp)+,d1
	rts

|***************************************|
| Display a string (A0) on the front panel display
|***************************************|
display_string:
	movem.l	d0-d2/a0-a1,-(sp)
	move.b	#CMD_POSITION,d1
	bra.s	ds.entry
ds.loop:
	move.b	d1,d0
	bsr.s	display_cmd		| select display position
	addq.b	#1,d1
ds.chop:
	and.w	#0x7f,d2
	sub.w	#0x20,d2
	bmi.s	ds.chop
	mulu.w	#7,d2
	lea	(font,pc),a1
	lea	(a1,d2.w),a1
	moveq	#7-1,d2
ds.loop2:
	move.b	(a1)+,d0
	bsr.s	display_cmd		| write 5 bits
	dbra	d2,ds.loop2
ds.entry:
	move.b	(a0)+,d2
	bne.s	ds.ok
	moveq	#' ',d2			| pad short string with spaces
	subq.l	#1,a0
ds.ok:
	cmp.b	#CMD_POSITION+8,d1
	blo.s	ds.loop
	movem.l	(sp)+,d0-d2/a0-a1
	rts

|======================================================================|
| Display font (5x7) for characters 0x20..0x7f
|======================================================================|
font:
	dc.b	0x00,0x00,0x00,0x00,0x00,0x00,0x00	|  
	dc.b	0x04,0x04,0x04,0x04,0x04,0x00,0x04	| !
	dc.b	0x0a,0x0a,0x00,0x00,0x00,0x00,0x00	| "
	dc.b	0x0a,0x0a,0x1f,0x0a,0x1f,0x0a,0x0a	| #
	dc.b	0x04,0x0f,0x14,0x0e,0x05,0x1e,0x04	| $
	dc.b	0x18,0x19,0x02,0x04,0x08,0x13,0x03	| %
	dc.b	0x08,0x14,0x14,0x08,0x15,0x12,0x0c	| &
	dc.b	0x0c,0x0c,0x04,0x08,0x00,0x00,0x00	| '
	dc.b	0x02,0x04,0x04,0x04,0x04,0x04,0x02	| (
	dc.b	0x08,0x04,0x04,0x04,0x04,0x04,0x08	| )
	dc.b	0x00,0x0a,0x04,0x1f,0x04,0x0a,0x00	| *
	dc.b	0x00,0x04,0x04,0x1f,0x04,0x04,0x00	| +
	dc.b	0x00,0x00,0x00,0x18,0x18,0x08,0x10	| ,
	dc.b	0x00,0x00,0x00,0x1f,0x00,0x00,0x00	| -
	dc.b	0x00,0x00,0x00,0x00,0x00,0x0c,0x0c	| .
	dc.b	0x00,0x01,0x02,0x04,0x08,0x10,0x00	| /
	dc.b	0x0e,0x11,0x11,0x11,0x11,0x11,0x0e	| 0
	dc.b	0x04,0x0c,0x04,0x04,0x04,0x04,0x0e	| 1
	dc.b	0x0e,0x11,0x01,0x06,0x08,0x10,0x1f	| 2
	dc.b	0x0e,0x11,0x01,0x0e,0x01,0x11,0x0e	| 3
	dc.b	0x02,0x06,0x0a,0x12,0x1f,0x02,0x02	| 4
	dc.b	0x1f,0x10,0x1e,0x01,0x01,0x01,0x1e	| 5
	dc.b	0x06,0x08,0x10,0x1e,0x11,0x11,0x0e	| 6
	dc.b	0x1f,0x01,0x02,0x04,0x08,0x08,0x08	| 7
	dc.b	0x0e,0x11,0x11,0x0e,0x11,0x11,0x0e	| 8
	dc.b	0x0e,0x11,0x11,0x0f,0x01,0x02,0x0c	| 9
	dc.b	0x00,0x0c,0x0c,0x00,0x0c,0x0c,0x00	| :
	dc.b	0x0c,0x0c,0x00,0x0c,0x0c,0x04,0x08	| ;
	dc.b	0x01,0x02,0x04,0x08,0x04,0x02,0x01	| <
	dc.b	0x00,0x00,0x1f,0x00,0x1f,0x00,0x00	| =
	dc.b	0x10,0x08,0x04,0x02,0x04,0x08,0x10	| >
	dc.b	0x0e,0x11,0x01,0x02,0x04,0x00,0x04	| ?
	dc.b	0x0e,0x11,0x17,0x15,0x17,0x10,0x0e	| @
	dc.b	0x0e,0x11,0x11,0x1f,0x11,0x11,0x11	| A
	dc.b	0x1e,0x11,0x11,0x1e,0x11,0x11,0x1e	| B
	dc.b	0x0e,0x11,0x10,0x10,0x10,0x11,0x0e	| C
	dc.b	0x1e,0x11,0x11,0x11,0x11,0x11,0x1e	| D
	dc.b	0x1f,0x10,0x10,0x1e,0x10,0x10,0x1f	| E
	dc.b	0x1f,0x10,0x10,0x1e,0x10,0x10,0x10	| F
	dc.b	0x0e,0x11,0x10,0x10,0x13,0x11,0x0e	| G
	dc.b	0x11,0x11,0x11,0x1f,0x11,0x11,0x11	| H
	dc.b	0x0e,0x04,0x04,0x04,0x04,0x04,0x0e	| I
	dc.b	0x01,0x01,0x01,0x01,0x01,0x11,0x0e	| J
	dc.b	0x11,0x12,0x14,0x18,0x14,0x12,0x11	| K
	dc.b	0x10,0x10,0x10,0x10,0x10,0x10,0x1f	| L
	dc.b	0x11,0x1b,0x15,0x15,0x11,0x11,0x11	| M
	dc.b	0x11,0x11,0x19,0x15,0x13,0x11,0x11	| N
	dc.b	0x0e,0x11,0x11,0x11,0x11,0x11,0x0e	| O
	dc.b	0x1e,0x11,0x11,0x1e,0x10,0x10,0x10	| P
	dc.b	0x0e,0x11,0x11,0x11,0x15,0x12,0x0d	| Q
	dc.b	0x1e,0x11,0x11,0x1e,0x14,0x12,0x11	| R
	dc.b	0x0e,0x11,0x10,0x0e,0x01,0x11,0x0e	| S
	dc.b	0x1f,0x04,0x04,0x04,0x04,0x04,0x04	| T
	dc.b	0x11,0x11,0x11,0x11,0x11,0x11,0x0e	| U
	dc.b	0x11,0x11,0x11,0x0a,0x0a,0x04,0x04	| V
	dc.b	0x11,0x11,0x11,0x15,0x15,0x1b,0x11	| W
	dc.b	0x11,0x11,0x0a,0x04,0x0a,0x11,0x11	| X
	dc.b	0x11,0x11,0x0a,0x04,0x04,0x04,0x04	| Y
	dc.b	0x1f,0x01,0x02,0x04,0x08,0x10,0x1f	| Z
	dc.b	0x07,0x04,0x04,0x04,0x04,0x04,0x07	| [
	dc.b	0x00,0x10,0x08,0x04,0x02,0x01,0x00	| \
	dc.b	0x1c,0x04,0x04,0x04,0x04,0x04,0x1c	| ]
	dc.b	0x04,0x0e,0x11,0x00,0x00,0x00,0x00	| ^
	dc.b	0x00,0x00,0x00,0x00,0x00,0x00,0x1f	| _
	dc.b	0x0c,0x0c,0x08,0x04,0x00,0x00,0x00	| `
	dc.b	0x00,0x00,0x0e,0x12,0x12,0x12,0x0d	| a
	dc.b	0x10,0x10,0x10,0x16,0x19,0x11,0x1e	| b
	dc.b	0x00,0x00,0x0e,0x10,0x10,0x11,0x0e	| c
	dc.b	0x01,0x01,0x01,0x0d,0x13,0x11,0x0f	| d
	dc.b	0x00,0x00,0x0e,0x11,0x1e,0x10,0x0e	| e
	dc.b	0x04,0x0a,0x08,0x1c,0x08,0x08,0x08	| f
	dc.b	0x00,0x00,0x0f,0x11,0x0f,0x01,0x06	| g
	dc.b	0x10,0x10,0x16,0x19,0x11,0x11,0x11	| h
	dc.b	0x00,0x04,0x00,0x0c,0x04,0x04,0x0e	| i
	dc.b	0x02,0x00,0x06,0x02,0x02,0x12,0x0c	| j
	dc.b	0x10,0x10,0x12,0x14,0x18,0x14,0x12	| k
	dc.b	0x0c,0x04,0x04,0x04,0x04,0x04,0x0e	| l
	dc.b	0x00,0x00,0x0a,0x15,0x11,0x11,0x11	| m
	dc.b	0x00,0x00,0x16,0x19,0x11,0x11,0x11	| n
	dc.b	0x00,0x00,0x0e,0x11,0x11,0x11,0x0e	| o
	dc.b	0x00,0x00,0x1e,0x11,0x19,0x16,0x10	| p
	dc.b	0x00,0x00,0x0f,0x11,0x13,0x0d,0x01	| q
	dc.b	0x00,0x00,0x0b,0x0c,0x08,0x08,0x08	| r
	dc.b	0x00,0x00,0x0e,0x10,0x0e,0x01,0x1e	| s
	dc.b	0x08,0x08,0x1c,0x08,0x08,0x0a,0x04	| t
	dc.b	0x00,0x00,0x09,0x09,0x09,0x0b,0x05	| u
	dc.b	0x00,0x00,0x11,0x11,0x11,0x0a,0x04	| v
	dc.b	0x00,0x00,0x11,0x11,0x15,0x15,0x0a	| w
	dc.b	0x00,0x00,0x11,0x0a,0x04,0x0a,0x11	| x
	dc.b	0x00,0x00,0x11,0x0a,0x04,0x04,0x08	| y
	dc.b	0x00,0x00,0x1f,0x02,0x04,0x08,0x1f	| z
	dc.b	0x02,0x04,0x04,0x08,0x04,0x04,0x02	| {
	dc.b	0x04,0x04,0x04,0x04,0x04,0x04,0x04	| |
	dc.b	0x08,0x04,0x04,0x02,0x04,0x04,0x08	| }
	dc.b	0x00,0x00,0x08,0x15,0x02,0x00,0x00	| ~
	dc.b	0x1f,0x11,0x11,0x11,0x11,0x11,0x1f	| ^?

|======================================================================|
| String constants and buffers
|======================================================================|

msg_flashing:
	.asciz	"Flashing"
msg_checking:
	.asciz	"Checking"
msg_crcok:
	.asciz	"CRC OK"
msg_writing:
	.asciz	"Writing"
msg_verify:
	.asciz	"Verify"
msg_flashok:
	.asciz	"Flash OK"
msg_reboot:
	.asciz	"Reboot.."
err_badcmd:
	.asciz	"Bad cmd!"
err_big:
	.asciz	"Too big!"
err_noalign:
	.asciz	"No align"
err_crc:
	.asciz	"Bad CRC!"
err_ram:
	.asciz	"RAM Err!"
err_failed:
	.asciz	"Failed!"
err_timeout:
	.asciz	"Timeout!"
err_notfound:
	.asciz	"NotFound"
	.align	2
_end:
crc_tab:
